import problemdetails

project = 'Tornado Problem Details'
copyright = '2019, Dave Shawley'
author = 'Dave Shawley'
release = problemdetails.version

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
]

html_theme = 'haiku'
intersphinx_mapping = {
    'https://docs.python.org/': None,
    'https://www.tornadoweb.org/en/stable/': None,
}



from tornado import httputil, web


class Problem(web.HTTPError):
    """An exception that will be translated into a json document.

    :param int status_code: HTTP status code to return
    :param str log_message: optional log message that is passed
        to the :class:`tornado.web.HTTPError` initializer
    :param args: parameters that are passed to `log_message` in
        the :class:`~tornado.web.HTTPError` initializer.
    :keyword str reason: optional reason phrase to use in the
        HTTP response line.  *This value is NOT included in the
        response document*.
    :param kwargs: additional keyword parameters are included in
        the response document

    :meth:`problemdetails.ErrorWriter.write_error` recognizes this
    exception type and renders :attr:`.document` as the *problem+json*
    result.  The *status* property is set to `status_code` and the
    *type* property will be set by ``write_error`` unless it is
    explicitly set.

    .. attribute:: document

       The keyword parameters are collected into this :class:`dict`
       and rendered as the response document

    """

    def __init__(self, status_code, log_message=None, *args, **kwargs):
        if status_code not in httputil.responses:
            kwargs.setdefault('reason', 'Abnormal Status')
        super(Problem, self).__init__(status_code, log_message, *args,
                                      **kwargs)
        self.document = kwargs
        self.document['status'] = status_code
        self.document.pop('reason', None)
        self.document.pop('log_message', None)



import json

from tornado import web

from problemdetails import Problem


def _rfc_link(rfc_no, anchor=None):
    if anchor:
        return 'https://tools.ietf.org/html/rfc{0}#{1}'.format(rfc_no, anchor)
    return 'https://tools.ietf.org/html/rfc{0}'.format(rfc_no)


type_link_map = {
    100: _rfc_link(7231, 'section-6.2.1'),
    101: _rfc_link(7231, 'section-6.2.2'),
    200: _rfc_link(7231, 'section-6.3.1'),
    201: _rfc_link(7231, 'section-6.3.2'),
    202: _rfc_link(7231, 'section-6.3.3'),
    203: _rfc_link(7231, 'section-6.3.4'),
    204: _rfc_link(7231, 'section-6.3.5'),
    205: _rfc_link(7231, 'section-6.3.6'),
    206: _rfc_link(7233, 'section-4.1'),
    207: _rfc_link(4918),
    208: _rfc_link(5842),
    226: _rfc_link(3229),
    300: _rfc_link(7231, 'section-6.4.1'),
    301: _rfc_link(7231, 'section-6.4.2'),
    302: _rfc_link(7231, 'section-6.4.3'),
    303: _rfc_link(7231, 'section-6.4.4'),
    304: _rfc_link(7232, 'section-4.1'),
    305: _rfc_link(7231, 'section-6.4.5'),
    306: _rfc_link(7231, 'section-6.4.6'),
    307: _rfc_link(7231, 'section-6.4.7'),
    308: _rfc_link(7538),
    400: _rfc_link(7231, 'section-6.5.1'),
    401: _rfc_link(7235, 'section-3.1'),
    402: _rfc_link(7231, 'section-6.5.2'),
    403: _rfc_link(7231, 'section-6.5.3'),
    404: _rfc_link(7231, 'section-6.5.4'),
    405: _rfc_link(7231, 'section-6.5.5'),
    406: _rfc_link(7231, 'section-6.5.6'),
    407: _rfc_link(7235, 'section-3.2'),
    408: _rfc_link(7231, 'section-6.5.7'),
    409: _rfc_link(7231, 'section-6.5.8'),
    410: _rfc_link(7231, 'section-6.5.9'),
    411: _rfc_link(7231, 'section-6.5.10'),
    412: _rfc_link(7232, 'section-4.2'),
    413: _rfc_link(7231, 'section-6.5.11'),
    414: _rfc_link(7231, 'section-6.5.12'),
    415: _rfc_link(7231, 'section-6.5.13'),
    416: _rfc_link(7233, 'section-4.4'),
    417: _rfc_link(7231, 'section-6.5.14'),
    418: _rfc_link(2324, 'section-2.3.2'),
    421: _rfc_link(7540, 'section-9.1.2'),
    422: _rfc_link(4918),
    423: _rfc_link(4918),
    424: _rfc_link(4918),
    425: _rfc_link(8470),
    426: _rfc_link(7231, 'section-6.5.15'),
    428: _rfc_link(6585),
    429: _rfc_link(6585),
    431: _rfc_link(6585),
    451: _rfc_link(7725),
    500: _rfc_link(7231, 'section-6.6.1'),
    501: _rfc_link(7231, 'section-6.6.2'),
    502: _rfc_link(7231, 'section-6.6.3'),
    503: _rfc_link(7231, 'section-6.6.4'),
    504: _rfc_link(7231, 'section-6.6.5'),
    505: _rfc_link(7231, 'section-6.6.6'),
    506: _rfc_link(2295),
    507: _rfc_link(4918),
    508: _rfc_link(5842),
    510: _rfc_link(2274),
    511: _rfc_link(6585),
}
"""Mapping of HTTP status code to *type* link.

This table maps HTTP status codes to the IANA registered
specification for the code.  You can add additional links or
replace ones that are here as you see fit.  The error writer
uses this table to generate the default ``type`` link in
responses.

"""


class ErrorWriter(web.RequestHandler):
    """Render *application/problem+json* in ``write_error``

    Include this class in the base class list to return errors as
    `application/problem+json`_ documents instead of HTML.

    .. attribute:: PROBLEM_DETAILS_MIME_TYPE

       This class-level attribute controls the content type used by
       `write_error`.  Due to the way that `write_error` is called,
       it is necessary to configure this.  Setting the content type
       header before calling :meth:`~tornado.web.RequestHandler.send_error`
       DOES NOT work as expected.

       If you need to change the response content type, change this
       attribute at the class-level.

    .. _application/problem+json: https://tools.ietf.org/html/rfc7807

    """

    PROBLEM_DETAILS_MIME_TYPE = "application/problem+json"

    def write_error(self, status_code, **kwargs):
        """Render *application/problem+json* documents instead of HTML.

        :param int status_code: HTTP status code that we returned
        :keyword str detail: optional *detail* field to include in
            the error document.  This field is omitted by default.
        :keyword str instance: optional *instance* to include in the
            error document.  This field is omitted by default.
        :keyword str title: optional *title* to include in the error
            document.  THis field is omitted by default.
        :keyword str type:  optional *type* field to include in the
            error document.  This field defaults to a link to the
            official HTTP specification of `status_code` if omitted
            and `status_code` is a standard code.

        See :rfc:`7807#section-3.1` for a description of each optional
        field.

        """
        body = {}
        if 'exc_info' in kwargs:
            exc_value = kwargs['exc_info'][1]
            if isinstance(exc_value, Problem):
                body = exc_value.document.copy()

        if not body:
            status_code = int(status_code)
            body = {'status': status_code}
            sentinel = object()
            for kwarg in ('detail', 'instance', 'title', 'type'):
                if kwargs.get(kwarg, sentinel) is not sentinel:
                    body[kwarg] = kwargs[kwarg]

        if 'type' not in body:
            try:
                body['type'] = type_link_map[status_code]
            except KeyError:
                pass

        self.set_header('Content-Type', self.PROBLEM_DETAILS_MIME_TYPE)
        self.write(json.dumps(body).encode('utf-8'))



"""Pair of handlers that implement the examples from :rfc:`7807`."""
import logging
import os

from tornado import ioloop, web
import problemdetails


class AccountHandler(problemdetails.ErrorWriter, web.RequestHandler):
    def post(self, account):
        price = float(self.get_query_argument('price'))
        self.verify_funds(account, price)
        self.buy_thing()

    def verify_funds(self, account, price):
        balance = self.get_balance(account)
        if price > balance:
            raise problemdetails.Problem(
                status_code=403,
                title='You do not have enough credit.',
                detail=(f'Your current balance is {balance}, but that '
                        f'costs {price}'),
                instance=self.reverse_url('account-handler', account),
                balance=balance,
                accounts=self.lookup_accounts(account),
                type='https://example.com/probs/out-of-credit',
            )

    def get_balance(self, account):
        return 30.0  # TODO look up balance

    def lookup_accounts(self, primary_account):
        related = [12345, 67890]  # TODO lookup related accounts
        return [self.reverse_url('account-handler', aid) for aid in related]

    def buy_thing(self):
        pass


class ValidationError(problemdetails.ErrorWriter, web.RequestHandler):
    def initialize(self):
        super(ValidationError, self).initialize()
        self.errors = []

    def get(self):
        age = self._extract_age()
        color = self._extract_color()
        if self.errors:
            raise problemdetails.Problem(**{
                'status_code': 400,
                'type': 'https://example.net/validation-error',
                'invalid-params': self.errors,
            })
        self.write({'age': age, 'color': color})

    def _record_error(self, field, error):
        self.errors.append({'name': field, 'reason': error})

    def _extract_age(self):
        try:
            age = self.get_query_argument('age')
        except web.MissingArgumentError:
            self._record_error('age', 'is required')
        else:
            try:
                age = int(age)
                if age > 0:
                    return age
                self._record_error('age', 'must be a positive integer')
            except ValueError:
                self._record_error('age', 'must be a integer')
        return None

    def _extract_color(self):
        try:
            color = self.get_query_argument('color')
        except web.MissingArgumentError:
            self._record_error('color', 'is required')
        else:
            if color in ('blue', 'green', 'red'):
                return color
            self._record_error('color', "must be 'green', 'red' or 'blue'")
        return None


if __name__ == '__main__':
    logging.basicConfig(
        level=logging.DEBUG, format='%(levelname)1.1s - %(name)s: %(message)s')
    app = web.Application([
        web.url(r'/account/(?P<account>\d+)', AccountHandler,
                name='account-handler'),
        web.url(r'/invalid-params', ValidationError),
    ], debug=True)

    iol = ioloop.IOLoop.current()
    port = int(os.environ.get('PORT', '8000'))
    stem = 'http://127.0.0.1:{0}'.format(port)
    logging.info('listening on %s', stem)
    logging.info('POST %s/account/1234?price=40', stem)
    logging.info('GET  %s/invalid-params', stem)
    logging.info('GET  %s/invalid-params?age=-10&color', stem)

    app.listen(address='127.0.0.1', port=port)
    try:
        iol.start()
    except KeyboardInterrupt:
        iol.stop()



try:
    from problemdetails.errors import Problem
    from problemdetails.handlers import ErrorWriter, type_link_map
except ImportError as error:  # pragma: no cover # noqa: 841

    class ErrorWriter:
        def __init__(self, *args, **kwargs):
            raise error  # noqa: 821

    class Problem(Exception):
        def __init__(self, *args, **kwargs):
            raise error  # noqa: 821


__all__ = [
    'ErrorWriter', 'Problem', 'type_link_map', 'version_info', 'version'
]

version_info = (0, 0, 5)
version = '.'.join(str(c) for c in version_info)



"""
Simple handler that invokes httpbin.org/status.

This file contains a simple handler that sends a request to httpbin.org,
receives an error response (from Tornado), and formats it into a problem
document.  The example handler shows how to catch a HTTPError and
transform it into a problemdetails.Problem.

"""
import logging
import os

from tornado import gen, httpclient, ioloop, web
import problemdetails


class HttpBinHandler(problemdetails.ErrorWriter, web.RequestHandler):
    @gen.coroutine
    def get(self):
        logger = logging.getLogger('HttpBinHandler')
        url = 'http://httpbin.org/status/{0}'.format(
            self.get_query_argument('status', '500'))
        logger.info('retrieving %s', url)

        client = httpclient.AsyncHTTPClient()
        try:
            response = yield client.fetch(url)
            self.add_header('Content-Type', 'application/json')
            self.write(response.body)
        except httpclient.HTTPError as error:
            raise problemdetails.Problem(
                status_code=error.code,
                httpbin_headers=dict(error.response.headers.items()))


if __name__ == '__main__':
    logging.basicConfig(
        level=logging.DEBUG, format='%(levelname)1.1s - %(name)s: %(message)s')
    app = web.Application([web.url('/', HttpBinHandler)], debug=True)

    iol = ioloop.IOLoop.current()
    port = int(os.environ.get('PORT', '8000'))
    stem = 'http://127.0.0.1:{0}'.format(port)
    logging.info('listening on %s', stem)
    logging.info('GET %s/?status=418', stem)

    app.listen(address='127.0.0.1', port=port)
    try:
        iol.start()
    except KeyboardInterrupt:
        iol.stop()



import json
try:
    from urllib.parse import urlencode
except ImportError:  # pragma: no cover
    from urllib import urlencode

from tornado import httputil, testing, web

from problemdetails import handlers
import problemdetails


class Application(web.Application):
    def __init__(self, **settings):
        super(Application, self).__init__([web.url('/', Handler)], settings)


class Handler(problemdetails.ErrorWriter, web.RequestHandler):
    def get(self):
        sentinel = object()  # safely detect missing query params

        status = int(self.get_query_argument('status', default='200'))
        raise_error = self.get_query_argument('raise_error', default=sentinel)

        if raise_error is not sentinel:
            query_args = [
                arg for arg in self.request.query_arguments.keys()
                if arg not in ('raise_error', 'status')
            ]
            kwargs = {}
            for name in query_args:
                kwargs[name] = self.get_query_argument(name)
                if kwargs[name].lower() == 'none':
                    kwargs[name] = None
                elif kwargs[name].startswith(('{', '[')):
                    kwargs[name] = json.loads(kwargs[name])
            raise problemdetails.Problem(status_code=status, **kwargs)
        else:
            kwargs = {}
            for name in ('detail', 'instance', 'title', 'type'):
                value = self.get_query_argument(name, sentinel)
                if value is not sentinel:
                    if value.lower() == 'none':
                        value = None
                    kwargs[name] = value
            if status not in httputil.responses:
                kwargs['reason'] = 'Abnormal Status'
            self.send_error(status, **kwargs)


class ErrorWriterTests(testing.AsyncHTTPTestCase):
    def setUp(self):
        super(ErrorWriterTests, self).setUp()
        self._original_content_type = (
            handlers.ErrorWriter.PROBLEM_DETAILS_MIME_TYPE)

    def tearDown(self):
        super(ErrorWriterTests, self).tearDown()
        handlers.ErrorWriter.PROBLEM_DETAILS_MIME_TYPE = (
            self._original_content_type)

    def get_app(self):
        return Application()

    def send_query(self, **query):
        return self.fetch('/?{0}'.format(urlencode(query)))

    def test_that_send_error_sets_content_type(self):
        response = self.send_query(status=500)
        self.assertEqual(response.code, 500)
        self.assertEqual(response.headers['Content-Type'],
                         'application/problem+json')

    def test_that_send_error_includes_status(self):
        response = self.send_query(status=500)
        body = json.loads(response.body.decode('utf-8'))
        self.assertEqual(body['status'], 500)

    def test_that_send_error_sets_type_to_rfc_url(self):
        response = self.send_query(status=500)
        body = json.loads(response.body.decode('utf-8'))
        self.assertEqual(body['type'],
                         handlers.type_link_map[response.code])
        self.assertEqual(body['type'],
                         'https://tools.ietf.org/html/rfc7231#section-6.6.1')

    def test_that_type_is_not_set_for_unknown_status_code(self):
        response = self.send_query(status=600)
        self.assertEqual(response.code, 600)
        body = json.loads(response.body.decode('utf-8'))
        self.assertNotIn('type', body)

    def test_that_type_can_be_overridden(self):
        response = self.send_query(
            status=500, type='http://example.com/errors#500')
        body = json.loads(response.body.decode('utf-8'))
        self.assertEqual(body['type'], 'http://example.com/errors#500')

    def test_that_title_can_be_set(self):
        response = self.send_query(status=500, title='Uh oh')
        body = json.loads(response.body.decode('utf-8'))
        self.assertEqual(body['title'], 'Uh oh')

    def test_that_instance_can_be_set(self):
        response = self.send_query(
            status=500, instance='http://example.com/probs/out-of-credit')
        body = json.loads(response.body.decode('utf-8'))
        self.assertEqual(body['instance'],
                         'http://example.com/probs/out-of-credit')

    def test_that_detail_can_be_set(self):
        response = self.send_query(status=500, detail='something bad happened')
        body = json.loads(response.body.decode('utf-8'))
        self.assertEqual(body['detail'], 'something bad happened')

    def test_that_instance_can_be_null(self):
        response = self.send_query(status=500, instance=None)
        body = json.loads(response.body.decode('utf-8'))
        self.assertIsNone(body['instance'], repr(body))

    def test_that_content_type_can_be_set(self):
        handlers.ErrorWriter.PROBLEM_DETAILS_MIME_TYPE = 'application/json'
        response = self.send_query()
        self.assertEqual(response.headers['Content-Type'], 'application/json')


class ProblemTests(ErrorWriterTests):
    def get_app(self):
        return Application()

    def send_query(self, **query):
        query['raise_error'] = True
        return self.fetch('/?{0}'.format(urlencode(query)))

    def test_that_custom_attributes_can_be_set(self):
        response = self.send_query(
            status=401, custom=json.dumps({'a': ['list']}))
        body = json.loads(response.body.decode('utf-8'))
        self.assertEqual(body['custom'], {'a': ['list']}, repr(body['custom']))

    def test_that_http_reason_added_for_nonstandard_response_codes(self):
        response = self.send_query(status=600)
        self.assertEqual(response.code, 600)
        self.assertEqual(response.reason, 'Abnormal Status')

    def test_that_reason_param_is_dropped_from_body(self):
        response = self.send_query(status=401, reason='whatever')
        body = json.loads(response.body.decode('utf-8'))
        self.assertNotIn('reason', body)

    def test_that_log_message_param_is_dropped_from_body(self):
        response = self.send_query(status=401, log_message='whatever')
        body = json.loads(response.body.decode('utf-8'))
        self.assertNotIn('log_message', body)



#!/usr/bin/env python

import setuptools

import problemdetails

setuptools.setup(
    name='tornado-problem-details',
    version=problemdetails.version,
    description='RFC-7807 Error documents for Tornado',
    long_description=open('README.rst').read(),
    author='Dave Shawley',
    author_email='daveshawley@gmail.com',
    url='https://github.com/dave-shawley/tornado-problem-details',
    packages=['problemdetails'],
    install_requires=[
        'tornado>=4.4',
    ],
    extras_require={
        'dev': [
            'coverage==4.5.3',
            'flake8==3.7.7',
            'flake8-fixme==1.1.0',
            'flake8-print==3.1.0',
            'nose==1.3.7',
            'readme-renderer==24.0',
            'sphinx==2.0.0',
            'tox==3.8.4',
            'twine==1.13.0',
            'wheel==0.33.1',
            'yapf==0.26.0',
        ],
        'docs': [
            'sphinx==2.0.0',
        ],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
        'Topic :: Software Development :: Libraries',
    ],
)
